package org.zeroturnaround.javarebel.integration.spring;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.zeroturnaround.javarebel.IntegrationFactory;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;
import org.zeroturnaround.javarebel.integration.spring.cbp.PathMatchingResourcePatternResolverCBP;

/**
 * @author Rein Raudjärv
 * 
 * @see PathMatchingResourcePatternResolver
 * @see PathMatchingResourcePatternResolverCBP
 */
public class PathMatchingResourcePatternResolverUtil {
  
  private static final Logger log = LoggerFactory.getInstance();
  
  private static class SubPathInfo {
    public int count;               // count of duplicates
    public boolean foundByRebel;    // at least one duplicate found by JavaRebel
  }
  
  public static Resource[] filter(PathMatchingResourcePatternResolver resolver, String rootDirPath, Resource[] rootDirResources, Resource[] result) {
    try {
      return filterInternal(resolver, rootDirPath, rootDirResources, result);
    } catch (Throwable t) {
      log.error(t);
      return result; // return original result
    }
  }
  
  private static Resource[] filterInternal(PathMatchingResourcePatternResolver resolver, String rootDirPath, Resource[] rootDirResources, Resource[] result) {
    if (!rootDirPath.startsWith(ResourcePatternResolver.CLASSPATH_ALL_URL_PREFIX))
      return result;
    rootDirPath = rootDirPath.substring(ResourcePatternResolver.CLASSPATH_ALL_URL_PREFIX.length());
    
    List list = new ArrayList(Arrays.asList(result));
    
    /*
     * Find root directory URLS (all of them and URLs found by JavaRebel)
     */
    Set rootDirURLs = getURLs(rootDirResources);
    Set rebelRootDirURLs = getRebelURLs(resolver.getClassLoader(), rootDirPath);
    
    /*
     * Collect data about sub paths
     */
    Map infos = new HashMap();
    for (Iterator it = list.iterator(); it.hasNext();) {
      Resource resource = (Resource) it.next();
      
      String url = getURL(resource);
      if (url == null)
        continue;
      
      for (Iterator rootDirIt = rootDirURLs.iterator(); rootDirIt.hasNext();) {
        String rootURL = (String) rootDirIt.next();
        if (!url.startsWith(rootURL))
          continue;
        
        String subPath = url.substring(rootURL.length());
        boolean foundByRebel = rebelRootDirURLs.contains(rootURL);
        
        SubPathInfo info = (SubPathInfo) infos.get(subPath);
        if (info == null) {
          info = new SubPathInfo();
          infos.put(subPath, info);
        }
        
        info.count++;
        
        if (foundByRebel)
          info.foundByRebel = true;
      }
    }
    
    /*
     * Remove duplicates (only if exactly 2 instances found)
     * by excluding native URLs replaced by JavaRebel
     */
    for (Iterator it = list.iterator(); it.hasNext();) {
      Resource resource = (Resource) it.next();
      String url = getURL(resource);
      if (url == null)
        continue;
      
      for (Iterator rootDirIt = rootDirURLs.iterator(); rootDirIt.hasNext();) {
        String rootURL = (String) rootDirIt.next();
        if (!url.startsWith(rootURL))
          continue;
      
        String subPath = url.substring(rootURL.length());
        boolean foundByRebel = rebelRootDirURLs.contains(rootURL);
        
        SubPathInfo info = (SubPathInfo) infos.get(subPath);
        
        /*
         * Remove sub path duplication not found by JavaRebel
         * if the same sub path is found by JavaRebel
         * and there are exactly 2 duplicates with this name.
         */
        if (info.count == 2 && info.foundByRebel && !foundByRebel) {
          it.remove();
          
          if (log.isEnabled())
            log.log("Excluding " + resource);
        }
      }
    }

    result = (Resource[]) list.toArray(new Resource[list.size()]);
    return result;
  }
  
  private static String getURL(Resource resource) {
    String url = null;
    try {
      url = resource.getURL().toExternalForm();
    } catch (IOException e) {
    }
    return url;
  }
  
  private static Set getURLs(Resource[] resources) {
    Set result = new LinkedHashSet();
    for (int i = 0; i < resources.length; i++) {
      Resource resource = resources[i];
      String url = getURL(resource);
      if (url != null)
        result.add(url);
    }
    return result;
  }
  
  private static Set getRebelURLs(ClassLoader loader, String path) {
    Set result = new LinkedHashSet();
    Enumeration en = IntegrationFactory.getInstance().findRebelResources(loader, path);
    if (en != null) {
      while (en.hasMoreElements()) {
        URL url = (URL) en.nextElement();
        result.add(url.toExternalForm());
      }
    }
    return result;
  }

}
