package org.zeroturnaround.javarebel.integration.spring;

import org.zeroturnaround.javarebel.StopWatch;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.springframework.aop.framework.Advised;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.context.annotation.ComponentScanBeanDefinitionParser;
import org.zeroturnaround.javarebel.ClassEventListener;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * This is the central class of the JavaRebel Spring plugin. 
 * It is the entry point for both bean definition reloading ({@link #onClassEvent(int, Class)}) 
 * and bean reconfiguration ({@link #reloadBeanDefinitions()}.  
 * 
 * @author Jevgeni Kabanov
 */
public final class SpringBeanReloader implements ClassEventListener {
  private static final Logger log = LoggerFactory.getInstance();

  private static final boolean isAnnotationBeanDefinitionEnabled = isAnnotationBeanDefinitionEnabled();
  private static ThreadLocal changedBeans = new ThreadLocal();
  
  private ConfigurableListableBeanFactory bf;    

  public SpringBeanReloader(ConfigurableListableBeanFactory bf) {
    this.bf = bf;
  }

  /**
   * This method will reconfigure the singletons when their class changes. 
   * It will delay reconfiguration if we are running in a {@link #reloadBeanDefinitions()}
   * method to be sure we are reconfiguring against the latest definitions.
   */
  public void onClassEvent(int eventType, Class klass) {
    StopWatch sw = log.createStopWatch("SpringBeanReloader.onClassEvent");

    try {
      if (changedBeans.get() == null)
        reloadBeanDefinitions();

      Map beans = bf.getBeansOfType(klass);

      for (Iterator i = beans.entrySet().iterator(); i.hasNext(); ) {      
        Map.Entry beanEntry = (Map.Entry) i.next();
        String beanName = (String) beanEntry.getKey();
        Object bean = beanEntry.getValue();

        boolean hasValidDefinition = true;

        // Since Spring 2.0 FactoryBean#getObject() may also return null
        if (bean == null)
          continue;

        bean = unwrapAdvisedBean(bean);

        // Checking if the bean is the result of factory call
        try {
          BeanDefinition bdf = bf.getBeanDefinition(beanName);
          if (bdf.getBeanClassName().endsWith("FactoryBean")) {
            hasValidDefinition = false;     
          }
        }
        catch (NoSuchBeanDefinitionException e) {
          hasValidDefinition = false;
        }

        // Reconfiguring/autowiring bean
        if (klass.isAssignableFrom(bean.getClass())) {
          reprocessBean(beanName, bean, hasValidDefinition);
        }
      }
    }
    finally {
      if (sw != null) sw.stop();
    }
  }

  private void reprocessBean(final String beanName, final Object bean,
      final boolean hasValidDefinition) {
    StopWatch sw = log.createStopWatch("SpringBeanReloader.reprocessBean");

    try {
      Runnable run = new Runnable() {
        public void run() {
          if (hasValidDefinition) {
            log.echo("JavaRebel-Spring: Reconfiguring bean '" + beanName + "' [" + bean.getClass().getName() + "]");
            bf.configureBean(bean, beanName);
          }
          else {
            // Even factory created bean can be autowired
            log.echo("JavaRebel-Spring: Autowiring annotations for bean of type [" + bean.getClass().getName() + "]");
            bf.autowireBean(bean);
          }
        }
      };

      if (changedBeans.get() == null) {
        // Do not delay it          
        run.run();
      } else {
        // Delay the action         
        ((List) changedBeans.get()).add(run);
      }
    }
    finally {
      if (sw != null) sw.stop();
    }
  }

  private Object unwrapAdvisedBean(Object bean) {
    // Unwrapping Advised beans
    try {
      while (bean instanceof Advised) {
        bean = ((Advised) bean).getTargetSource().getTarget();
      }
    } catch (Exception e) {
      LoggerFactory.getInstance().error(e);
    }
    return bean;
  }

  public int priority() {
    return PRIORITY_CORE;
  }

  /**
   * This method will reload the bean definitions and apply postprocessing. 
   * As it touches classes it may cause them to reload if they were changed.
   */
  public static void reloadBeanDefinitions() {
    if (changedBeans.get() != null)
      return;
    
    StopWatch sw = log.createStopWatch("SpringBeanReloader.reloadBeanDefinitions");

    try {
      changedBeans.set(new ArrayList());
      try {
        XmlBeanDefinitionManager.checkAndReload();

        if (isAnnotationBeanDefinitionEnabled) {
          AnnotationBeanDefinitionManager.checkRescan();
        }
      }
      finally {
        List toRun = (List) changedBeans.get();
        changedBeans.set(null);

        for (Iterator i = toRun.iterator(); i.hasNext();) {
          Runnable run = (Runnable) i.next();
          run.run();
        }                 
      }
    }
    finally {
      if (sw != null) sw.stop();
    }
  }

  public static boolean isReloadingBeanDefinitions() {
    return changedBeans.get() != null;
  }

  private static long lastCheck = 0;

  public static void checkAndReloadBeanDefinitions() {
    if (lastCheck + 1000 > System.currentTimeMillis())
      return;    

    reloadBeanDefinitions();

    lastCheck = System.currentTimeMillis();
  }

  /**
   * @return <code>true</code> if AnnotationBeanDefinitionManager can be used.
   */
  private static boolean isAnnotationBeanDefinitionEnabled() {
    try {
      ComponentScanBeanDefinitionParser.class.getName();
      return true;
    } catch (NoClassDefFoundError e) {
      return false;
    }
  }
}